#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <time.h>

#define tol 1.e-10

void StefanProblem_TI_BoundaryCondition(int n, int NT, double h, double k, double t[], double **A, double **B, double **L, double **R, double *w0, double *w1, double *w, double *xi_i, double *b, double *h_num);
void StefanProblem_TD_BoundaryCondition(int n, int NT, double h, double k, double t[], double **A, double **B, double **L, double **R, double *w0, double *w1, double *w, double *xi_i, double *b, double *h_num);

double Find_z(int n, double beta, double h, double k, double z[], double *w0, double *w1, double *xi_i);
double Find_h(int n, int l, double beta, double h, double k, double z[], double *w0, double *w1, double *xi_i, double t[]);

double f(double x, double a[], double hv[]);
double df(double x, double a[], double hv[]);
double newton(double x0, double a[], double hv[]);

double *Solve_Tridiagonal_Crout(int n, double **A, double *b);

void Omplir_L(int n, double **L, double **U, double **A);
void Omplir_U(int n, double **U, double **A);

double *triang_inf(int n, double **A, double *F);
double *triang_sup(int n, double **A, double *F);

double *mult_MatVec_Tridiag(int n, double **A, double *x);

int main(void){
    clock_t time;
    /*Nombre de nodes*/
    int n=10;
    /*Nombre de passos temporals*/
    int NT = n;
    /*Pas espacial h*/
    double h=1./n;
    /*Pas temporal k*/
    double k=h;
    
    int i;
    double *b, *w, *w0, *w1, *xi_i, *h_num;
    double **A, **B, **L, **R;
    
    double t[NT+1];
    for(i=0; i<NT+1; i++){
        t[i] = i*k;
    }
    
    /*------------------Guardem memòria per les matrius i vectors------------------*/
    b = (double *) calloc(n+1, sizeof(double));
    w = (double *) calloc(n+1, sizeof(double));
    w0 = (double *) calloc(n+1, sizeof(double));
    w1 = (double *) calloc(n+1, sizeof(double));
    h_num = (double *) calloc(n+1, sizeof(double));
    xi_i = (double *) calloc(n+1, sizeof(double));
    if(b == NULL || w == NULL || w0 == NULL || w1 == NULL || h_num == NULL || xi_i == NULL){
        printf("No hi ha prou memòria\n");
        exit (1);
    }
    A = (double **) calloc(3, sizeof(double *));
    B = (double **) calloc(3, sizeof(double *));
    L = (double **) calloc(3, sizeof(double *));
    R = (double **) calloc(3, sizeof(double *));
    if(A == NULL || B == NULL || L == NULL || R == NULL){
        printf("No hi ha prou memòria\n");
        exit (1);
    }
    for(i=0; i<n+1; i++){
        A[i] = (double *) calloc(n+1, sizeof(double));
        B[i] = (double *) calloc(n+1, sizeof(double));
        L[i] = (double *) calloc(n+1, sizeof(double));
        R[i] = (double *) calloc(n+1, sizeof(double));
        if(A[i] == NULL || B[i] == NULL || L[i] == NULL || R[i] == NULL){
            printf("No hi ha prou memòria\n");
            exit (1);
        }
    }
    
    /*Omplim A:*/
    for(i=1; i<n; i++){
        A[0][i] = (1./(2*(h*h)))*(-2.);
        if(i<n-1) A[1][i] = (1./(2*(h*h)));
        if(i>1) A[2][i] = (1./(2*(h*h)));
    }
    
    /*Omplim el vector de posició \xi:*/
    for(i=0; i<n+1; i++){
        xi_i[i] = i*h;
    }
    
    time=clock();
    StefanProblem_TI_BoundaryCondition(n,NT,h,k,t,A,B,L,R,w0,w1,w,xi_i,b,h_num);
    time=clock() - time;
    double t_taken = ((double) time)/CLOCKS_PER_SEC;
    printf("TI, execution time = %16.8e\n", t_taken);
    time=clock();
    StefanProblem_TD_BoundaryCondition(n,NT,h,k,t,A,B,L,R,w0,w1,w,xi_i,b,h_num);
    time=clock() - time;
    t_taken = ((double) time)/CLOCKS_PER_SEC;
    printf("TD, execution time = %16.8e\n", t_taken);
    
    /*Alliberem memòria per a les matrius i els vectors*/
    free(b);
    free(w);
    free(w0);
    free(w1);
    free(xi_i);
    free(B[0]); free(B[1]); free(B[2]);
    free(L[0]); free(L[1]); free(L[2]);
    free(R[0]); free(R[1]); free(R[2]);
    
    return 0;
}

/*Resol problema de Stefan per a la condició a la frontera independent del temps f(t)=1*/
void StefanProblem_TI_BoundaryCondition(int n, int NT, double h, double k, double t[], double **A, double **B, double **L, double **R, double *w0, double *w1, double *w, double *xi_i, double *b, double *h_num){
    int i, l;
    double z_aux, z_err;
    double beta, lambda;
    
    /*Inicialitzem z_j i z_j+1*/
    double z[2];
    z[0] = 0;
    z[1] = 0;
    
    /*Obrim els fitxers d'entrada i sortida*/
    FILE *in1, *in2;                              
    in1 = fopen("lambda.in", "r");
    in2 = fopen("beta02.in", "r");
    if (in1 == NULL || in2 == NULL){
    printf("Error al obrir el fitxer d'entrada.\n");
    exit (-1);
    }
    
    fscanf(in1, "%lf", &lambda);
    fscanf(in2, "%lf", &beta);
    
    
    FILE *out1, *out2;
    out1 = fopen("freeBoundaryPos.out", "w");
    out2 = fopen("temperatureDistribution.out", "w");
    if (out1 == NULL || out2 == NULL){
        printf("Error al obrir el fixer de sortida.\n");
        exit (-1);
    }
    
    /*Omplim B:*/
    for(i=1; i<n; i++){
        B[0][i] = 0.;
        if(i<n-1) B[1][i] = (1./(8.*h))*xi_i[i];
        if(i>1) B[2][i] = -(1./(8.*h))*xi_i[i];
    }
    
    i=0;
    while(fscanf(in1, "%lf", &w[i]) != EOF){
        if(i>0 && i<n){
            w0[i] = w[i];
        }
        i++;
    }
    fclose(in1);
    
    h_num[0] = z[0];
    fprintf(out1, "%16.8e %16.8e\n", t[0], h_num[0]);
    for(i=0; i<n+1; i++){
        fprintf(out2, "%16.8e %16.8e\n", xi_i[i]*h_num[0], 1*w[i]);
    }
    fprintf(out2, "\n");
    
    for(l=1; l<NT+1; l++){
        z_err = 1e-12;
        while(z_err > 1e-13){
            double z_mid = (z[1]+z[0])/2;
            double dz = (z[1]-z[0])/k;
            
            for(i=1; i<n; i++){
                L[0][i] = A[0][i] - z_mid/k;
                if(i<n-1) L[1][i] = A[1][i] + dz*B[1][i];
                if(i>1) L[2][i] = A[2][i] + dz*B[2][i];
            }
            
            for(i=1; i<n; i++){
                R[0][i] = -A[0][i] - z_mid/k;
                if(i<n-1) R[1][i] = -A[1][i] - dz*B[1][i];
                if(i>1) R[2][i] = -A[2][i] - dz*B[2][i];
            }
            
            b = mult_MatVec_Tridiag(n, R, w0);
            
            double cte;
            cte = 1/(2*h*h)-(xi_i[1]/(8*h))*dz;
            b[1] = b[1] - 2*cte;
            
            /*Ressolució del sistema Lw1=Rw0 - 2cte */
            w1 = Solve_Tridiagonal_Crout(n, L, b);
            
            z_aux = z[1];
            z[1] = Find_z(n, beta, h, k, z, w0, w1, xi_i); 
            z_err = fabs(z[1] - z_aux);
        } 
        
        /*Condicions a la frontera*/
        w[0] = 1.;
        w[n] = 0.;
        for(i=1; i<n; i++){
            w[i] = w1[i];
        }
        
        /*Desfem els canvis de variable i guardem u(x,t) i h(t) en fitxers separats*/
        h_num[l] = sqrt(z[1]);
        fprintf(out1, "%16.8e %16.8e\n", t[l], h_num[l]);
        for(i=0; i<n+1; i++){
            fprintf(out2, "%16.8e %16.8e\n", xi_i[i]*h_num[l], 1*w[i]);
        }
        fprintf(out2, "\n");
        
        w0 = w1;
        z[0] = z[1];
        z[1] = z[1];
    }
    
    /*Tanquem els fitxers de sortida*/
    fclose(out1); fclose(out2);
}

/*Resol problema de Stefan per a la condició a la frontera dependent del temps f(t)=e^t-1*/
void StefanProblem_TD_BoundaryCondition(int n, int NT, double h, double k, double t[], double **A, double **B, double **L, double **R, double *w0, double *w1, double *w, double *xi_i, double *b, double *h_num){
    int i, l;
    double beta=1.;
    double lambda = 1./sqrt(beta);
    
    double hv_aux, hv_err;
    double hv[2];
    hv[0] = 0.;
    hv[1] = 0.;
    
    FILE *out1, *out2, *out1_anal, *out1_err, *out2_anal, *out2_err;
    out1 = fopen("freeBoundaryPosTD.out", "w");
    out2 = fopen("temperatureDistributionTD.out", "w");
    out1_anal = fopen("freeBoundaryPosTD_anal.out", "w");
    out1_err = fopen("freeBoundaryPosTD_err.out", "w");
    out2_anal = fopen("temperatureDistributionTD_anal.out", "w");
    out2_err = fopen("temperatureDistributionTD_err.out", "w");
    if (out1 == NULL || out2 == NULL || out1_anal == NULL || out1_err == NULL || out2_anal == NULL || out2_err == NULL){
        printf("Error al obrir el fixer de sortida.\n");
        exit (-1);
    }
    
    /*Omplim B:*/
    for(i=1; i<n; i++){
        B[0][i] = 0.;
        if(i<n-1) B[1][i] = (1./(4.*h))*xi_i[i];
        if(i>1) B[2][i] = -(1./(4.*h))*xi_i[i];
    }
    
    for(i=0; i<n+1; i++){
        w[i] = 1 - xi_i[i];
        if(i>0 && i<n){
            w0[i] = w[i];
        }
    }
    
    h_num[0] = hv[0];
    fprintf(out1, "%16.8e %16.8e\n", t[0], h_num[0]);
    double h_anal = lambda*t[0];
    fprintf(out1_anal, "%16.8e %16.8e\n", t[0], h_anal);
    fprintf(out1_err, "%16.8e %16.8e\n", t[0], fabs(h_num[0]-h_anal));
        
    double u_num, u_anal;
    double sum=0.;
    for(i=0; i<n+1; i++){
        u_num = t[0]*w[i];
        fprintf(out2, "%16.8e %16.8e\n", xi_i[i]*h_num[0], u_num);
        u_anal = exp(t[0]-xi_i[i]*h_anal)-1.;
        fprintf(out2_anal, "%16.8e %16.8e\n", xi_i[i]*h_num[0], u_anal);
        sum += pow(u_anal - u_num, 2);
    }
    fprintf(out2, "\n");
    fprintf(out2_anal, "\n");
    fprintf(out2_err, "%16.8e %16.8e\n", t[0], sqrt(h*sum));

    for(l=1; l<NT+1; l++){
        hv_err = 1e-12;
        
        double y;
        if(l==1) y = (exp(t[l])-1)/t[l] + 1;
        else y = (exp(t[l])-1)/t[l] + (exp(t[l-1])-1)/t[l-1];
        
        while(hv_err > 1e-13){
            double hv_mid = (hv[1]+hv[0])/2;
            double dhv = (hv[1]-hv[0])/k;
            double t_mid = (t[l]+t[l-1])/2;
            
            for(i=1; i<n; i++){
                L[0][i] = t_mid*A[0][i] - pow(hv_mid,2.)*(1./2. + t_mid/k);
                if(i<n-1) L[1][i] = t_mid*A[1][i] + t_mid*hv_mid*dhv*B[1][i];
                if(i>1) L[2][i] = t_mid*A[2][i] + t_mid*hv_mid*dhv*B[2][i];
            }
            
            for(i=1; i<n; i++){
                R[0][i] = -t_mid*A[0][i] + pow(hv_mid,2.)*(1./2. - t_mid/k);
                if(i<n-1) R[1][i] = -t_mid*A[1][i] - t_mid*hv_mid*dhv*B[1][i];
                if(i>1) R[2][i] = -t_mid*A[2][i] - t_mid*hv_mid*dhv*B[2][i];
            }
            
            b = mult_MatVec_Tridiag(n, R, w0);
            
            double d;
            d = (-t_mid*(1./(2.*pow(h,2.))) + xi_i[1]*t_mid*hv_mid*dhv*(1./(4.*h)))*y;
            b[1] = b[1] + d;
            
            w1 = Solve_Tridiagonal_Crout(n, L, b);
            
            hv_aux = hv[1];
            hv[1] = Find_h(n, l, beta, h, k, hv, w0, w1, xi_i, t); 
            hv_err = fabs(hv[1] - hv_aux);
        }
        
        if(l==1){
            w[0] = 1.;
        }
        else {
            w[0] = y - w[0];
        }
        w[n] = 0.;
        for(i=1; i<n; i++){
            w[i] = w1[i];
        }
        
        h_num[l] = hv[1];
        fprintf(out1, "%16.8e %16.8e\n", t[l], h_num[l]);
        h_anal = lambda*t[l];
        fprintf(out1_anal, "%16.8e %16.8e\n", t[l], h_anal);
        fprintf(out1_err, "%16.8e %16.8e\n", t[l], fabs(h_num[l]-h_anal));
        
        sum=0.;
        for(i=0; i<n+1; i++){
            u_num = t[l]*w[i];
            fprintf(out2, "%16.8e %16.8e\n", xi_i[i]*h_num[l], u_num);
            u_anal = exp(t[l]-xi_i[i]*h_anal)-1.;
            fprintf(out2_anal, "%16.8e %16.8e\n", xi_i[i]*h_num[l], u_anal);
            sum += pow(u_anal - u_num, 2);
        }
        fprintf(out2, "\n");
        fprintf(out2_anal, "\n");
        fprintf(out2_err, "%16.8e %16.8e\n", t[l], sqrt(h*sum));
        
        w0 = w1;
        hv[0] = hv[1];
        hv[1] = hv[1];        
    }
    
    /*Tanquem els fitxers de sortida*/
    fclose(out1); fclose(out2); 
    fclose(out1_anal); fclose(out1_err); 
    fclose(out2_anal); fclose(out2_err);
}

/*Trobem z resolent la equació quadràtica*/
double Find_z(int n, double beta, double h, double k, double z[], double *w0, double *w1, double *xi_i){
    double a, b, c;
    double z1, z2;
    double r, v;
    
    r = k/(h*h);
    v = k/h;
    
    /*Coeficients de l'equació quadràtica per a z_{j+1}*/
    a = beta*xi_i[n]/(2*k);
    b = -beta*(xi_i[n]/k)*z[0]+2*beta*r/v;
    c = beta*(xi_i[n]/(2*k))*(z[0]*z[0])-2*(beta*r/v)*z[0]-2*r*(w1[n-1]+w0[n-1]);

    /*Arrels de z[1]*/
    z1 = (-b+sqrt(b*b-4*a*c))/(2*a); 
    z2 = (-b-sqrt(b*b-4*a*c))/(2*a);
    
    /*Agafem la solució positiva*/
    if(z1 > z2) z[1] = z1;
    else z[1] = z2;
    
    return z[1];
}

/*Trombem h resolent l'equació quàrtica amb el mètode de Newton-Raphson*/
double Find_h(int n, int l, double beta, double h, double k, double hv[], double *w0, double *w1, double *xi_i, double t[]){
    double r, v;
    
    r = k/(h*h);
    v = k/h;

    double a[4];
    a[0] = 2*r*(t[l]+t[l-1])*(w1[n-1]+w0[n-1]);
    a[1] = 4*beta/v;
    a[2] = r;
    a[3] = xi_i[n]/(4.*h);

    return newton(hv[1], a, hv);
}

double f(double x, double a[], double hv[]){
    return a[0] - a[1]*(a[2]+a[3]*(pow(x,2)-pow(hv[0],2)))*(pow(x,2)-pow(hv[0],2));
}

double df(double x, double a[], double hv[]){
    double h=1.e-5;
    return (f(x+h,a,hv)-f(x,a,hv))/h;
}

/*Mètode de Newton per trobar h*/
double newton(double x0, double a[], double hv[]){
    double temp = -1.;
    double x1;
    while (fabs(temp-x0)>1e-10) {
	x1 = x0 - f(x0, a, hv)/df(x0, a, hv);
	temp = x0;
	x0 = x1;
    };
    return x1;
}

/*Ressolució sistema tridiagonal Aw=b*/
double *Solve_Tridiagonal_Crout(int n, double **A, double *b){
    int i;
    double *aux, *w;
    double **L, **U;

    w = (double *) calloc(n+1, sizeof(double));
    aux = (double *) calloc(n+1, sizeof(double));
    if(aux == NULL || w == NULL){
        printf("No hi ha prou memòria\n");
        exit (1);
    }
    
    L = (double **) calloc(2, sizeof(double *));
    U = (double **) calloc(2, sizeof(double *));
    if (L == NULL || U == NULL){
        printf("Error de memòria.\n");
        exit(1);
    }
    for(i=0; i<2; i++){
        L[i] = (double *) calloc(n+2, sizeof(double));
        U[i] = (double *) calloc(n+2, sizeof(double));
        if (L[i] == NULL || U[i] == NULL){
            printf("Error de memòria.\n");
            exit(1);
        }
    }
    
    /*Descomposició LU*/
    Omplir_U(n,U,A);
    Omplir_L(n,L,U,A);
    
    /*Resolem sistema Lx=b i Uy=x*/
    aux = triang_inf(n,L,b);
    w = triang_sup(n,U,aux);
    
    /*Alliberem memòria*/
    free(aux);
    free(L[0]); free(L[1]);
    free(U[0]); free(U[1]);
    
    return w;
}

void Omplir_L(int n, double **L, double **U, double **A){
    int i;
    for(i=1; i<n; i++){
        L[0][i] = 1.;
        if(i>1) L[1][i] = A[2][i]/U[0][i-1];
    }
    return;
}

void Omplir_U(int n, double **U, double **A){
    int i;
    U[0][1] = A[0][1];
    U[1][1] = A[1][1];
    for(i=2; i<n; i++){
        U[0][i] = A[0][i] - (A[1][i-1]*A[2][i])/U[0][i-1];
        if(i<n-1) U[1][i] = A[1][i];
    }
    return;
}

/*Resol sistema triangular inferior, A matriu tridiagonal*/
double *triang_inf(int n, double **A, double *F){
    int i;
    double *x;
    x = (double *) calloc(n+2, sizeof(double));
    if (x==NULL){
        printf("Error de memòria.\n");
    }
    
    x[1] = (1./A[0][1])*F[1];
    for(i=2; i<n; i++){
        x[i] = (1./A[0][i])*(F[i]-A[1][i]*x[i-1]);
    }
    
    return x;
}

/*Resol sistema triangular superior, A matriu tridiagonal*/
double *triang_sup(int n, double **A, double *F){
    int i;
    double *x;
    x = (double *) calloc(n+2, sizeof(double));
    if (x==NULL){
        printf("Error de memòria.\n");
    }
    
    x[n-1] = (1./A[0][n-1])*F[n-1];
    for(i=n-2; i>=1; i--){
        x[i] = (1./A[0][i])*(F[i]-A[1][i]*x[i+1]);
    }
    
    return x;
}

/*Multiplicació matriu tridiagonal A per vector x: Ax=y*/
double *mult_MatVec_Tridiag(int n, double **A, double *x){
    int i;
    double *y;
    y = (double *) calloc(n+1, sizeof(double));
    if (y==NULL){
        printf("Error de memòria.\n");
    } 
    y[1] = A[0][1]*x[1] + A[1][1]*x[2];
    for(i=2; i<n-1; i++){
        y[i] = A[2][i]*x[i-1] + A[0][i]*x[i] + A[1][i]*x[i+1];
    }
    y[n-1] = A[2][n-1]*x[n-2] + A[0][n-1]*x[n-1];
    
    return y;
}
